import matplotlib.pyplot as plt
import numpy as np

from lambast.generate_data import HSMM


def run_example():
    """
    Example for the HSMM class
    """
    # Define number of time series to generate and their length
    n, t = 5, 100  # Generate 5 time series of length 100

    # Define HSMM parameters

    # Initial state probabilities
    init_probs = [0.7, 0.3]

    transition_probs = [[0.8, 0.2], [0.3, 0.7]]  # Transition probabilities

    emission_means = [[0, 0], [3, 3]]  # Mean vectors for states 0 and 1

    # Covariance matrices for state 0 and 1
    emission_covariances = [[[1, 0.2], [0.2, 1]],
                            [[1, -0.3], [-0.3, 1]]]

    # Duration parameters for each state
    state_durations_params = [
        (10, 3, 1, t),  # State 1: mean=10, std=3, min=5, max=t
        (20, 5, 1, t),  # State 2: mean=20, std=5, min=10, max=t
    ]

    # Initialize the HSMM
    rng = np.random.default_rng(seed=42)
    hsmm = HSMM(init_probs, transition_probs, emission_means,
                emission_covariances, state_durations_params, rng)

    # Generate time series data using above HSMM parameters
    samples, states = hsmm.sample(n, t)

    # Plot
    fig, axes = plt.subplots(3, 1, figsize=(10, 15))
    # Define a color palette for the time series
    colors = plt.cm.viridis(np.linspace(0, 1, n))

    def do_plot(all_data, title, xlabel="Time",
                ylabel="Value", label="Time Series"):
        """
        Factor out the plots
        """
        for i, (data, color) in enumerate(zip(all_data, colors)):
            axes[ax_index].plot(data, label=f"{label} {i + 1}", color=color)
        axes[ax_index].set_title(title)
        axes[ax_index].set_xlabel(xlabel)
        axes[ax_index].set_ylabel(ylabel)
        axes[ax_index].legend()

    # Transpose samples to access them easier for plot
    t_samples = np.transpose(samples)

    # Plot dimension 1 for all time series
    ax_index = 0
    do_plot(t_samples[0].T, "Dimension 1 of Time Series",
            label="Time Series")

    # Plot dimension 2 for all time series
    ax_index += 1
    do_plot(t_samples[1].T, "Dimension 2 of Time Series",
            label="Time Series")

    # Plot state sequences for all time series
    ax_index += 1
    do_plot(states, "State Sequences", ylabel="State",
            label="State Sequence")

    # Adjust layout and show the plot
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    run_example()



"""
lambast/mitigation_methods/density_ratio/train_task.py

Training utilities for binary classification task.

"""


from __future__ import annotations

from typing import Dict, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset

from .util import _to_float_tensor

TensorLike = Union[np.ndarray, torch.Tensor]


@torch.no_grad()
def eval_binary_accuracy(
    model: nn.Module,
    X: TensorLike,
    y: TensorLike,
    *,
    device: torch.device,
    batch_size: int = 512,
) -> Dict[str, float]:
    model.eval()
    X_t = _to_float_tensor(X)
    y_t = _to_float_tensor(y).view(-1)

    ds = TensorDataset(X_t, y_t)
    dl = DataLoader(ds, batch_size=batch_size, shuffle=False)

    correct = 0.0
    total = 0.0
    loss_sum = 0.0
    bce = nn.BCEWithLogitsLoss(reduction="sum")

    for xb, yb in dl:
        xb = xb.to(device)
        yb = yb.to(device)
        logits = model(xb)
        loss_sum += float(bce(logits, yb).item())
        preds = (torch.sigmoid(logits) > 0.5).float()
        correct += float((preds == yb).float().sum().item())
        total += float(yb.numel())

    return {
        "acc": correct / max(total, 1.0),
        "bce": loss_sum / max(total, 1.0),
    }


def train_binary_classifier(
    model: nn.Module,
    X: TensorLike,
    y: TensorLike,
    *,
    X_val=None,
    y_val=None,
    sample_weight: Optional[TensorLike] = None,
    device: Optional[torch.device] = None,
    epochs: int = 10,
    batch_size: int = 256,
    lr: float = 1e-3,
    weight_decay: float = 1e-4,
) -> Tuple[nn.Module, dict[str, list[float]]]:
    """
    Train a binary classifier on (X, y), optionally with per-example weights.

    X: (N, C, T)
    y: (N,) in {0,1}
    sample_weight: (N,) positive weights (optional)
    """
    device = device or torch.device(
        "cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    model.train()

    X_t = _to_float_tensor(X)
    y_t = _to_float_tensor(y).view(-1)

    if sample_weight is None:
        ds = TensorDataset(X_t, y_t)
    else:
        w_t = _to_float_tensor(sample_weight).view(-1)
        if w_t.shape[0] != X_t.shape[0]:
            raise ValueError("sample_weight must have same length as X.")
        ds = TensorDataset(X_t, y_t, w_t)

    dl = DataLoader(ds, batch_size=batch_size, shuffle=True, drop_last=False)
    has_val = (X_val is not None) and (y_val is not None)
    if has_val:
        Xv = _to_float_tensor(X_val)
        yv = _to_float_tensor(y_val).view(-1)
        val_ds = TensorDataset(Xv, yv)
        val_dl = DataLoader(val_ds, batch_size=batch_size, shuffle=False)
    else:
        val_dl = None

    opt = optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
    bce_none = nn.BCEWithLogitsLoss(reduction="none")

    history: Dict[str, list] = {
        "train_loss": [],
        "train_acc": [],
        "val_loss": [],
        "val_acc": [],
    }

    for _ in range(epochs):
        loss_sum = 0.0
        correct = 0.0
        total = 0.0
        for batch in dl:
            if sample_weight is None:
                xb, yb = batch
                wb = None
            else:
                xb, yb, wb = batch

            xb = xb.to(device)
            yb = yb.to(device)

            opt.zero_grad()
            logits = model(xb)
            per_ex_loss = bce_none(logits, yb)

            if wb is not None:
                wb = wb.to(device)
                loss = (per_ex_loss * wb).mean()
            else:
                loss = per_ex_loss.mean()

            loss.backward()
            opt.step()

            loss_sum += float(per_ex_loss.sum().item())
            preds = (torch.sigmoid(logits) > 0.5).float()
            correct += float((preds == yb).float().sum().item())
            total += float(yb.numel())
        train_loss = loss_sum / max(total, 1.0)
        train_acc = correct / max(total, 1.0)
        if val_dl is not None:
            val_loss, val_acc = _eval_binary_epoch(model, val_dl, device)
        else:
            val_loss, val_acc = float("nan"), float("nan")

        history["train_loss"].append(train_loss)
        history["train_acc"].append(train_acc)
        history["val_loss"].append(val_loss)
        history["val_acc"].append(val_acc)

    return model, history


@torch.no_grad()
def _eval_binary_epoch(
    model: nn.Module,
    dl: DataLoader,
    device: torch.device,
) -> Tuple[float, float]:
    """
    Returns (mean_loss, accuracy) over the dataloader.
    Loss is unweighted BCE (for evaluation).
    """
    model.eval()
    bce = nn.BCEWithLogitsLoss(reduction="sum")

    loss_sum = 0.0
    correct = 0.0
    total = 0.0

    for batch in dl:
        if len(batch) == 2:
            xb, yb = batch
        else:
            xb, yb, _ = batch  # ignore weights for eval

        xb = xb.to(device)
        yb = yb.to(device).view(-1)

        logits = model(xb)
        loss_sum += float(bce(logits, yb).item())

        preds = (torch.sigmoid(logits) > 0.5).float()
        correct += float((preds == yb).float().sum().item())
        total += float(yb.numel())

    mean_loss = loss_sum / max(total, 1.0)
    acc = correct / max(total, 1.0)

    model.train()
    return mean_loss, acc



from typing import Union

import numpy as np
import torch

TensorLike = Union[np.ndarray, torch.Tensor]


def _to_float_tensor(x: TensorLike) -> torch.Tensor:
    if isinstance(x, torch.Tensor):
        t = x
    else:
        t = torch.from_numpy(np.asarray(x))
    return t.float().contiguous()



"""
Time series generation classes.

"""

import copy
import warnings

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from scipy.stats import truncnorm
from typing_extensions import Self

from lambast.utils import assert_valid_covariance

from .timeseries_class import TimeSeries


class LinearSSM(TimeSeries):

    def __init__(self, state_matrix: NDArray, state_noise_cov: NDArray,
                 obs_matrix: NDArray, obs_noise_cov: NDArray,
                 rng: np.random.Generator | None = None,
                 scale_matrix: bool = False) -> None:
        """
        Initializes a new instance of a Linear State Space Model.
        State dimension is d, observation dimension is p.
        Assume zero-mean Gaussian noise in state and observation space.

        Parameters:
            state_matrix: shape (d,d) numpy array defining state
                transition matrix
            state_noise_cov: shape (d,d) numpy array defining
                state noise covariance
            obs_matrix: shape (p,d) numpy array defining
                observation matrix
            obs_noise_cov: shape (p,p) numpy array defininig
                observation noise covariance
            rng: numpy rng (else use default); see
                https://numpy.org/doc/2.0/reference/random/index.html#random-quick-start
            scale_matrix (bool): Whether to rescale the matrix for stability,
                default, false.
        """
        super().__init__()

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        self.d = state_matrix.shape[0]
        self.p = obs_matrix.shape[0]

        self.state_matrix = state_matrix
        self.state_noise_cov = state_noise_cov
        self.obs_matrix = obs_matrix
        self.obs_noise_cov = obs_noise_cov

        # Check if state matrix is stable and warn user if not, but only if
        # scale_matrix is false, if scale_matrix is true, just scale the matrix
        self.rescale_matrix(scale_matrix)

        # Check if covariance matrices are valid
        assert_valid_covariance(state_noise_cov)
        assert_valid_covariance(obs_noise_cov)

        # Check shapes
        if state_matrix.shape[0] != state_matrix.shape[1]:
            raise ValueError("State matrix is not square")
        if obs_matrix.shape[1] != self.d:
            raise ValueError("Obs matrix is not (p,d)")
        if state_noise_cov.shape[0] != self.d:
            raise ValueError("State cov matrix is not (d,d)")
        if obs_noise_cov.shape[0] != self.p:
            raise ValueError("Obs cov matrix is not (d,d)")

    def rescale_matrix(self, scale_matrix: bool) -> None:
        """
        Function that re-scales the state matrix to make it stable

        Parameter:
            scale_matrix: Whether to rescale the state matrix
        """
        eigenvalues = np.linalg.eigvals(self.state_matrix)
        spectral_radius = np.max(np.abs(eigenvalues))
        if spectral_radius >= 1:
            if scale_matrix:
                self.state_matrix /= (spectral_radius + 0.1)
            else:
                warnings.warn('Warning: state_matrix is not stable.')

    def copy_with_changes(self, **kwargs) -> Self:
        """
        Copy the initial parameters of this object into another object. Allow
        kwargs to change the initial values of the copy object.
        """

        # Copy current object
        other = copy.deepcopy(self)

        # Copy changed arguments
        for k in kwargs:
            other.__dict__[k] = kwargs[k]

        scale_key = "scale_matrix"
        if scale_key in kwargs:
            other.rescale_matrix(kwargs[scale_key])

        return other

    def evolve_state(self, state: NDArray) -> NDArray:
        """
        Docstring TODO
        """
        mv_n = self.rng.multivariate_normal
        draws = mv_n(np.zeros((self.d)), self.state_noise_cov,
                     size=(state.shape[0]))[:, :, np.newaxis]

        return self.state_matrix @ state + draws

    def get_obs(self, state: NDArray) -> NDArray:
        """
        Docstring TODO
        """
        mv_n = self.rng.multivariate_normal
        draws = mv_n(np.zeros((self.p)), self.obs_noise_cov,
                     size=(state.shape[0]))[:, :, np.newaxis]

        return self.obs_matrix @ state + draws

    def sample(self, n: int, t: int, init_mean: NDArray | None = None,
               init_cov: NDArray | None = None) -> NDArray:
        """
        Samples from LinearSSM with initial state sampled from Gaussian with
        init_mean and init_cov.
        Default: mean zero with unit variance.

        Parameters:
            n: number of time series replicates to sample
            t: number of time points in each time series
            init_mean: shape (d,1) mean values for initial state sampling
            init_cov: shape (d,1) covariance for initial state sampling

        Returns:
            Numpy array of shape (n,p,t) representing generated time series
        """
        self.ts_samples = np.zeros((n, self.p, t))
        if init_mean is None:
            init_mean = np.zeros((self.d))
        if init_cov is None:
            init_cov = np.eye(self.d)

        # Sample initial state
        state = self.rng.multivariate_normal(init_mean, init_cov,
                                             size=n)[:, :, np.newaxis]

        # Recursively sample observations
        for t_index in range(t):
            self.ts_samples[:, :, t_index] = self.get_obs(state)[:, :, 0]
            state = self.evolve_state(state)

        return self.ts_samples

    def plot_sample(self) -> None:
        """
        Plot the time series sample
        """
        plt.figure()

        for i in range(self.p):
            subplot_index = int(f"{self.p}1{i + 1}")
            plt.subplot(subplot_index)

            plt.plot(self.ts_samples[:, i, :].T)
            plt.title(f"Dim. {i}")

        plt.xlabel('Time')
        plt.show()


class HSMM(TimeSeries):

    def __init__(self, init_probs: NDArray, transition_probs: NDArray,
                 emission_means: list[NDArray],
                 emission_covariances: list[NDArray],
                 state_durations_params: list[tuple],
                 rng: np.random.Generator | None = None) -> None:
        """
        Initialize the Hidden Semi-Markov Model with multivariate emissions.

        Parameters:
            init_probs: A 1D numpy array of initial state probabilities
                (must sum to 1).
            transition_probs: A 2D numpy array (N x N) where N is the number of
                states. Each entry represents the probability of transitioning
                from one state to another.
            emission_means: A list of mean vectors (1D numpy arrays) for
                emissions in each state.
            emission_covariances: A list of covariance matrices
                (2D numpy arrays) for emissions in each state.
            state_durations_params: A list of tuples
                [(mean, std, min, max), ...] representing the mean, standard
                deviation, and truncation range for the duration of each state.
            rng: numpy rng (else use default); see
                https://numpy.org/doc/2.0/reference/random/index.html#random-quick-start
        """
        super().__init__()

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        self.init_probs = np.array(init_probs)
        self.transition_probs = np.array(transition_probs)
        self.emission_means = [np.array(mean) for mean in emission_means]
        self.emission_covariances = [np.array(cov) for cov in
                                     emission_covariances]
        self.state_durations_params = state_durations_params

        # Validate inputs
        num_states = len(self.init_probs)
        if not np.isclose(np.sum(self.init_probs), 1):
            raise ValueError("Initial state probabilities must sum to 1.")
        if self.transition_probs.shape[0] != self.transition_probs.shape[1]:
            raise ValueError("Transition matrix must be square.")
        if self.transition_probs.shape[0] != num_states:
            e = "Transition matrix dimensions must match the number of states."
            raise ValueError(e)
        if not np.allclose(self.transition_probs.sum(axis=1), 1):
            e = "Each row of the transition matrix must sum to 1."
            raise ValueError(e)
        if len(self.emission_means) != num_states:
            e = "Number of emission means must match the number of states."
            raise ValueError(e)
        if len(self.emission_covariances) != num_states:
            e = "Number of emission covariances must match the "
            e += "number of states."
            raise ValueError(e)
        if len(self.state_durations_params) != num_states:
            e = "Number of state duration parameters must match the "
            e += "number of states."
            raise ValueError(e)

        for cov in self.emission_covariances:
            if cov.shape[0] != cov.shape[1]:
                raise ValueError("Each covariance matrix must be square.")
            if not np.allclose(cov, cov.T):
                raise ValueError("Each covariance matrix must be symmetric.")
            if not np.all(np.linalg.eigvals(cov) >= 0):
                e = "Each covariance matrix must be positive semi-definite."
                raise ValueError(e)

    def truncated_discrete_normal(self, mean: float, std: float,
                                  min_val: float, max_val: float,
                                  size: int = 1) -> NDArray:
        """
        Sample from a truncated discrete normal distribution.

        Parameters:
            mean: Mean of the normal distribution.
            std: Standard deviation of the normal distribution.
            min_val: Minimum truncation value.
            max_val: Maximum truncation value.
            size: Number of samples to generate.

        Returns:
            A NumPy array of samples.
        """
        a = (min_val - mean) / std
        b = (max_val - mean) / std
        samples = truncnorm(a, b, loc=mean, scale=std).rvs(
            size=size, random_state=self.rng)

        return np.clip(np.round(samples), min_val, max_val).astype(int)

    def sample(self, n: int, t: int) -> tuple[list[NDArray], list[NDArray]]:
        """
        Generate n time series of length t from the Hidden Semi-Markov Model
        with multivariate emissions.

        Parameters:
            n: Number of time series to generate.
            t: Length of each time series.

        Returns:
            samples: A list of n numpy arrays, each of shape (t, D),
                representing the multivariate time series.
            states: A list of n numpy arrays, each of shape (t,), representing
                the hidden state sequence.
        """

        num_states = len(self.emission_means)
        samples = []
        states = []

        for _ in range(n):
            sequence: list[NDArray] = []
            state_sequence = []

            # Start in a random initial state
            current_state = self.rng.choice(num_states, p=self.init_probs)
            time = 0

            while time < t:
                # Sample duration from truncated discrete normal distribution
                values = self.state_durations_params[current_state]
                mean, std, min_val, max_val = values
                duration = self.truncated_discrete_normal(mean, std, min_val,
                                                          max_val, size=1)[0]

                # Limit duration to avoid exceeding the time series length
                duration = min(duration, t - time)

                # Generate multivariate emissions for the duration
                em = self.emission_means[current_state]
                ec = self.emission_covariances[current_state]
                emissions = self.rng.multivariate_normal(em, ec, duration)
                sequence.extend(emissions)
                state_sequence.extend([current_state] * duration)

                # Transition to the next state
                tp = self.transition_probs[current_state]
                next_state = self.rng.choice(num_states, p=tp)
                current_state = next_state

                time += duration

            # Append the generated sequence and states
            samples.append(np.array(sequence))
            states.append(np.array(state_sequence))

        return samples, states



from numpy.typing import NDArray


class TimeSeries(object):

    def __init__(self) -> None:
        """
        TimeSeries base class.

        """

        return None

    def sample(self, n: int, t: int, *args, **kwargs
               ) -> NDArray | tuple[list[NDArray], list[NDArray]]:
        """
        Sample `n` time series with `t` time points from the TimeSeries model.

        Parameters:
            n (int): number of time series replicates to sample
            t (int): number of time points in each time series
        """
        raise NotImplementedError("Subclasses must implement this method.")



"""
Density Ratio Weighting Demo: Seismic Binary Classification

This example demonstrates covariate shift mitigation using
density-ratio weighting on a synthetic seismic dataset.
The dataset is located at: https://zenodo.org/records/17943530

Setup:
    1. Place dataset files under:
       lambast/datasets/seistask/

    2. Expected files:
       - SeisTask_data.h5
       - SeisTask_metadata.csv

Here, we split source and target by source_type (gabor or ricker).

This script:
    - Trains a baseline classifier on source
    - Evaluates on target
    - Learns density-ratio weights using unlabeled source/target
    - Retrains weighted classifier
    - Compares performance

Run:
    python examples/density_ratio_seismic_binary.py
"""
# %%
from importlib import resources

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from sklearn.model_selection import train_test_split

from lambast.mitigation_methods.density_ratio import DensityRatioWeighter
from lambast.mitigation_methods.density_ratio.models import BinaryCNN
from lambast.mitigation_methods.density_ratio.train_task import (
    eval_binary_accuracy, train_binary_classifier)

np.random.seed(42)
torch.manual_seed(42)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Get path to lambast package
with resources.as_file(resources.files("lambast")) as package_root:
    data_dir = package_root / "datasets" / "seistask"

# %% Load data
with h5py.File('%s/SeisTask_data.h5' % data_dir, 'r') as f:
    # List all groups and datasets in the file
    print("Keys: %s" % f.keys())
    print(f['data'].shape)
    data = f['data'][:]

# split source/target and make train/test
meta = pd.read_csv('%s/SeisTask_metadata.csv' % data_dir)
source_ix = np.where(meta['source_type'] == 'ricker')[0]
target_ix = np.where(meta['source_type'] == 'gabor')[0]
source_y = meta.iloc[source_ix]['signal'].to_numpy()
target_y = meta.iloc[target_ix]['signal'].to_numpy()
source_x = data[source_ix]
target_x = data[target_ix]


def train_test_val(x, y, train_size=0.8):
    train_ix, val_ix = train_test_split(
        np.arange(len(x)), test_size=1 - train_size)
    val_ix, test_ix = train_test_split(val_ix, test_size=0.5)
    return {'x': {'train': x[train_ix], 'val': x[val_ix], 'test': x[test_ix]},
            'y': {'train': y[train_ix], 'val': y[val_ix], 'test': y[test_ix]}}


source_split = train_test_val(source_x, source_y)
target_split = train_test_val(target_x, target_y)

x_mean = np.mean(source_split['x']['train'], axis=0, keepdims=True)
x_std = np.std(source_split['x']['train'], axis=0, keepdims=True)
for k in ['train', 'val', 'test']:
    source_split['x'][k] = (source_split['x'][k] - x_mean) / x_std
    target_split['x'][k] = (target_split['x'][k] - x_mean) / x_std

plt.figure(figsize=(8, 6))
plt.subplot(211)
plt.plot(source_split['x']['train'][:10, 0, :].T)
plt.title('Source')
plt.subplot(212)
plt.plot(target_split['x']['train'][:10, 0, :].T)
plt.title('Target')
plt.show()

acc_results = []

# %% Baseline
# Train classifier on source data, test on target data
baseline_cnn = BinaryCNN(in_channels=2)
baseline, hist = train_binary_classifier(baseline_cnn,
                                         source_split['x']['train'],
                                         source_split['y']['train'],
                                         X_val=source_split['x']['val'],
                                         y_val=source_split['y']['val'],
                                         device=device,
                                         epochs=150, lr=0.01)
src_acc = eval_binary_accuracy(
    baseline,
    source_split['x']['test'],
    source_split['y']['test'],
    device=device)
trg_acc = eval_binary_accuracy(
    baseline,
    target_split['x']['test'],
    target_split['y']['test'],
    device=device)
print("Baseline (train on source) test accuracy on source: "
      f"{src_acc['acc']:0.2f}, target: {trg_acc['acc']:0.2f}")

plt.figure()
plt.plot(hist["train_loss"], label='train')
plt.plot(hist["val_loss"], label='val')
plt.legend()
plt.title('Baseline CNN loss')
plt.show()
acc_results.append(
    {'model': 'base_source', 'test_set': 'source', 'acc': src_acc['acc']})
acc_results.append(
    {'model': 'base_source', 'test_set': 'target', 'acc': trg_acc['acc']})

# train classifier on target data, test on target data
baseline_target_cnn = BinaryCNN(in_channels=2)
baseline_target, hist = train_binary_classifier(baseline_target_cnn,
                                                target_split['x']['train'],
                                                target_split['y']['train'],
                                                X_val=target_split['x']['val'],
                                                y_val=target_split['y']['val'],
                                                device=device,
                                                epochs=150,
                                                lr=0.01)
src_acc = eval_binary_accuracy(
    baseline_target,
    source_split['x']['test'],
    source_split['y']['test'],
    device=device)
trg_acc = eval_binary_accuracy(
    baseline_target,
    target_split['x']['test'],
    target_split['y']['test'],
    device=device)
print("Baseline (train on source) test accuracy on source: "
      f"{src_acc['acc']:0.2f}, target: {trg_acc['acc']:0.2f}")
plt.figure()
plt.plot(hist["train_loss"], label='train')
plt.plot(hist["val_loss"], label='val')
plt.legend()
plt.title('Baseline (target) CNN loss')
plt.show()
acc_results.append(
    {'model': 'base_target', 'test_set': 'source', 'acc': src_acc['acc']})
acc_results.append(
    {'model': 'base_target', 'test_set': 'target', 'acc': trg_acc['acc']})

# %% Density ratio weighting
# Train domain classifier to reweight source data
drw = DensityRatioWeighter(epochs=300, batch_size=256, device=device, lr=0.001)
drw = drw.fit(
    source_split['x']['train'],
    target_split['x']['train'],
    source_split['x']['val'],
    target_split['x']['val'])
print(drw.diagnostics_)

plt.figure()
plt.plot(drw.history_["train_loss"], label='train')
plt.plot(drw.history_["val_loss"], label='val')
plt.legend()
plt.title('Domain classifier loss')
plt.show()

w_source = drw.compute_weights(source_split['x']['train'], alpha=0.3)
print('weight mean: %0.2f, sd: %0.2f, min: %0.2f, max: %0.2f' %
      (w_source.mean(), w_source.std(), w_source.min(), w_source.max()))
ess = w_source.sum()**2 / (w_source**2).sum()
print('ESS/n %0.2f' % (ess / len(w_source)))


# %% Weighted training
# Weighted training: use density ratio weights and apply to target
weighted_cnn = BinaryCNN(in_channels=2)
weighted, hist = train_binary_classifier(weighted_cnn,
                                         source_split['x']['train'],
                                         source_split['y']['train'],
                                         sample_weight=w_source,
                                         X_val=source_split['x']['val'],
                                         y_val=source_split['y']['val'],
                                         device=device,
                                         epochs=150,
                                         lr=0.01)
src_acc = eval_binary_accuracy(
    weighted,
    source_split['x']['test'],
    source_split['y']['test'],
    device=device)
trg_acc = eval_binary_accuracy(
    weighted,
    target_split['x']['test'],
    target_split['y']['test'],
    device=device)
print("Weighted test accuracy on source: "
      f"{src_acc['acc']:0.2f}, target: {trg_acc['acc']:0.2f}")
plt.figure()
plt.plot(hist["train_loss"], label='train')
plt.plot(hist["val_loss"], label='val')
plt.legend()
plt.title('Weighted CNN loss')
plt.show()
acc_results.append(
    {'model': 'weighted', 'test_set': 'source', 'acc': src_acc['acc']})
acc_results.append(
    {'model': 'weighted', 'test_set': 'target', 'acc': trg_acc['acc']})
# %%
acc_results = pd.DataFrame(acc_results)
print(acc_results)



import warnings
from dataclasses import dataclass
from itertools import cycle, pairwise
from typing import Any, Callable, ClassVar

import matplotlib.pyplot as plt
import numpy as np
import ruptures
from numpy.typing import ArrayLike


@dataclass
class ChangePoint(object):
    """
    Class to detect changepoints
    """

    samples: ArrayLike | None = None
    detected: bool = False
    optimal_breakpoints: dict | None = None
    models: ClassVar[list[str]] = ["l1", "l2", "ar", "clinear", "cosine",
                                   "cosine", "linear", "mahalanobis", "normal",
                                   "rank", "rbf"]
    kernels: ClassVar[list[str]] = ["linear", "rbf", "cosine"]

    estimators: ClassVar[dict[str, Callable]] = {
        "Binseg": ruptures.Binseg,
        "BottomUp": ruptures.BottomUp,
        "Dynp": ruptures.Dynp,
        "KernelCPD": ruptures.KernelCPD,
        "Pelt": ruptures.Pelt,
        "Window": ruptures.Window
    }

    optional_args: ClassVar[list[str]] = [
        "model",
        "custom_cost",
        "min_size",
        "jump",
    ]

    choose_arg: ClassVar[dict[str, list[str]]] = {
        "Binseg": ["n_bkps", "pen", "epsilon"],
        "BottomUp": ["n_bkps", "pen", "epsilon"],
        "Dynp": ["n_bkps"],
        "KernelCPD": ["n_bkps", "pen"],
        "Pelt": ["pen"],
        "Window": ["n_bkps", "pen", "epsilon"],
    }

    def was_detected(self) -> bool:
        """
        Returns True if a changepoint was detected, False otherwise
        """

        return self.detected

    def set_samples(self, samples: ArrayLike) -> None:

        self.samples = np.array(samples)

    def get_estimators(self) -> str:
        """
        Retreive all possible estimators, with use parameters
        """

        s = "Estimators: "
        for key in ChangePoint.estimators:
            s += f"{key}: use with one of {ChangePoint.choose_arg[key]}\n"

        s += "--- Description of parameters ---\n"
        s += "n_bkps (int): number of breakpoints to find before stopping.\n"
        s += "penalty (float): penalty value (>0)\n"
        s += "epsilon (float): reconstruction budget (>0)"

        return s

    def get_models(self) -> str:
        """
        Retreive all possible models
        """

        s = f"Models: {ChangePoint.models}"
        return s

    def get_kernels(self) -> str:
        """
        Retreive all possible kernels
        """

        s = f"Kernels: {ChangePoint.kernels}"
        return s

    def detect_change_point(self, samples: ArrayLike | None = None,
                            estimator: str = "Pelt",
                            model: str = "rbf",
                            kernel: str = "rbf",
                            custom_cost: ruptures.base.BaseCost | None = None,
                            min_size: int = 2,
                            jump: int = 5,
                            **fit_args: dict[str, Any]) -> None:
        """
        Detect the change points

        Parameters:
        -samples: timeseries with potential changepoint
        -estimator: which estimator to use, choose one from get_estimators()
        -model: which cost function to use, choose one from get_models()
        -kernel: which kernel to use, choose one from get_kernels()
        -custom_cost: Custom cost function. Defaults to None.
        -min_size: Minimum segment length.
        -jump: subsample (one every *jump* points).
        -fit_args: a dictionary of parameters for the cost instance.
        """

        if samples is not None:
            self.samples = np.array(samples)
        assert type(self.samples) is np.ndarray

        self.optimal_breakpoints = None
        self.detected = False

        if self.samples is None or self.samples.shape[0] == 0:
            warnings.warn("No data, skipping detection")
            return

        est_args: dict[str, Any] = {"min_size": min_size, "jump": jump}
        if estimator == "KernelCPD":
            est_args["kernel"] = kernel
        else:
            est_args["model"] = model
            est_args["custom_cost"] = custom_cost

        self.optimal_breakpoints = ChangePoint.estimators[estimator](
            **est_args).fit(self.samples).predict(**fit_args)

        if len(self.optimal_breakpoints) > 1:
            self.detected = True

    def display(self, x_values: list[float] | None = None,
                save_fig: bool = True,
                figname: str = "test", filetype: str = "pdf",
                use_ax: plt.Axes | None = None,
                **plot_args) -> None:
        """
        Display the change points

        Optional parameters:
        -save_fig: whether to save the plotted figure to a file
        -figname: filename to save the plot, ignored if save_fig is false
        -filetype: savefile extension, ignored if save_fig is false
        -use_ax: specific plt.Axes to use, if used, assuming that figure will
            be displayed or saved elsewhere, so save_fig is ignored.
        """

        color_cycle = cycle(["#4286f4", "#f44174"])

        if self.samples is None:
            warnings.warn("No data to display")
            return

        if use_ax is None:
            if self.optimal_breakpoints is None:
                if x_values is None:
                    plt.plot(self.samples, **plot_args)
                else:
                    plt.plot(self.samples)
            else:
                ruptures.display(self.samples, self.optimal_breakpoints)
        else:
            if x_values is None:
                use_ax.plot(self.samples, **plot_args)
            else:
                use_ax.plot(x_values, self.samples, **plot_args)

            # Plot the breakpoints
            if self.optimal_breakpoints is not None:

                # NOTE:
                # This replicates the ruptures plotting of breakpoints
                # but allows us to use it in a multi-panel figure.
                bkps = [0] + list(self.optimal_breakpoints)
                for (start, end), col in zip(pairwise(bkps), color_cycle):
                    use_ax.axvspan(max(0, start - 0.5), end -
                                   0.5, facecolor=col, alpha=0.2)

        if use_ax is None and save_fig:
            plt.savefig(figname + '.' + filetype)



"""
Generate complex-valued NQR signals using Voigt model.

Authors: Natalie Klein, Amber Day
"""

from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.special import wofz

from lambast.utils import util


@dataclass
class Voigt(object):
    """
    Class to manage VoigtSignal and associated functions

    Inputs
        sample_n : number of examples
        nt : number of time points
        fs : sampling rate
        freq_range : upper and lower bounds for the frequency, freq
        phi_range : upper and lower bounds for the phase offset, phi
        decay_rate_rage: upper and lower bounds for the decay_rate
        sigma_range : upper and lower bounds for the sigma
        amp_range : upper and lower bounds for the initial
            amplitude, amp
        noise_var : variance for white noise generation
        seed: random seed
        f_vec: frequency vector for freqs
        t : corresponding time vector for sigs
    """

    sample_n: int = 10000
    nt: int = 1024
    fs: float = 1.0 / 1.8e-05
    freq_range: tuple[float, float] = (-50, 50)
    phi_range: tuple[float, float] = (-np.pi, np.pi)
    decay_rate_range: tuple[float, float] = (1e-3, 1e-2)
    sigma_range: tuple[float, float] = (1e-3, 1e-2)
    amp_range: tuple[float, float] = (0.1, 1)
    noise_var: float = 1
    rng: np.random.Generator = np.random.default_rng()

    def synthetic_data_gen(self) -> tuple[dict[str, Any], dict[str, Any]]:
        """
        Generating synthetic data with white noise

        Outputs
            df_in: df containing N, nt, fs, w, phi, T2, A, sigma,
                s, C, t, f_vec
            df: df containing sigs, noise, noisy_sigs, freqs
            sigs: time series voigt signals
            noise: time series white noise
            freqs: frequency series voigts signals
        """

        # Initializing Parameters
        freq = self.rng.uniform(
            self.freq_range[0], self.freq_range[1], self.sample_n)
        phi = self.rng.uniform(
            self.phi_range[0], self.phi_range[1], self.sample_n)
        decay_rate = self.rng.uniform(
            self.decay_rate_range[0], self.decay_rate_range[1], self.sample_n)
        sigma = self.rng.uniform(
            self.sigma_range[0], self.sigma_range[1], self.sample_n)
        amp = self.rng.uniform(
            self.amp_range[0], self.amp_range[1], self.sample_n)
        const = np.zeros(self.sample_n)
        df = pd.DataFrame({"freq": freq, "phi": phi, "decay_rate": decay_rate,
                          "amp": amp, "sigma": sigma, "const": const})

        # Signal Generation
        t = np.linspace(0, self.nt / self.fs, self.nt)
        f_vec = np.linspace(-500, 500, self.nt)
        sigs = self.sig_gen(t, df)
        freqs = self.freq_gen(f_vec, df)

        # Generate White Noise
        noise = util.white_noise_gen(
            t, self.sample_n, sigma=self.noise_var, rng=self.rng)
        snrs = util.compute_snr(sigs, noise)
        noisy_sig = sigs + noise
        df_in = {
            "N": self.sample_n,
            "nt": self.nt,
            "fs": self.fs,
            "freq": freq,
            "phi": phi,
            "decay_rate": decay_rate,
            "amp": amp,
            "sigma": sigma,
            "noise_var": self.noise_var,
            "const": const,
            "t": t,
            "f_vec": f_vec,
        }

        df_out = {
            "snr": snrs,
            "sigs": sigs,
            "noise": noise,
            "noisy_sig": noisy_sig,
            "freqs": freqs,
        }

        return df_in, df_out

    def __freq_or_sig_gen(self, vector: NDArray[np.float64],
                          df: pd.DataFrame, mode: str
                          ) -> NDArray[np.float64] | NDArray[np.complex64]:
        """
        Function factoring out the common parts of sig_gen and freq_gen
        Args:
            t: vector of time or frequency points
            df: Pandas data frame containing signal parameter columns amp,
                freq, decay_rate, phi, sigma, const, with N rows
            mode: "time" or "freq", for the type of return
        Returns:
            np.ndarray of signals, shape (N, len(vector))
        """

        assert mode in ["time", "freq"]
        dtype: type[complex] | type[float]

        if mode == "time":
            dtype = complex
        elif mode == "freq":
            dtype = float

        result = np.zeros((df.shape[0], len(vector)), dtype=dtype)
        for i in range(df.shape[0]):
            signal = VoigtSignal(df.freq.iloc[i], df.decay_rate.iloc[i],
                                 df.amp.iloc[i], df.phi.iloc[i],
                                 df.sigma.iloc[i], df.const.iloc[i])
            if mode == "time":
                result[i, :] = signal.time_signal(vector)
            elif mode == "freq":
                result[i, :] = signal.freq_signal(vector)

        return result

    def sig_gen(self, t: NDArray, df: pd.DataFrame) -> NDArray:
        """
        Generate collection of Voigt time series signals based on Pandas data
        frame of parameter values.
        Args:
            t: vector of time points
            df: Pandas data frame containing signal parameter columns amp,
                freq, decay_rate, phi, sigma, const, with N rows
        Returns:
            np.ndarray of signals, shape (N, len(t))
        """

        return self.__freq_or_sig_gen(t, df, mode="time")

    def freq_gen(self, f_vec: NDArray, df: pd.DataFrame) -> NDArray:
        """
        Generate collection of Voigt frequency signals based on Pandas
        data frame of parameter values.
        Args:
            f_vec: vector of frequency points
            df: Pandas data frame containing signal parameter columns amp,
                freq, decay_rate, phi, sigma, const, with N rows
        Returns:
            np.ndarray of signals, shape (N, len(f_vec))
        """

        return self.__freq_or_sig_gen(f_vec, df, mode="freq")


@dataclass
class VoigtSignal(object):
    """
    Voigt signal model. If sigma=np.inf, recover FID signal model.
    Args:
        freq: frequency (Hz)
        decay_rate: time decay constant
        amp: amplitude constant
        phi: phase shift constant
        sigma: shape constant
        const: constant shift
    """

    freq: float
    decay_rate: float
    amp: float
    phi: float
    sigma: float
    const: float

    def time_signal(self, t: NDArray[np.float64]) -> NDArray[np.complex64]:
        """
        Return complex-valued time-domain signal given time vector t.
        """

        exp1 = np.exp(-0.5 * (t / self.sigma) ** 2)
        exp2 = np.exp(-t / self.decay_rate)
        exp3 = np.exp(1j * (2 * np.pi * self.freq * t + self.phi))

        return self.amp * exp1 * exp2 * exp3 + self.const

    def freq_signal(self, f_vec: NDArray[np.float64]) -> NDArray[np.float64]:
        """
        Return real-valued frequency-domain representation given
        frequency vector f_vec.
        """

        exp_arg = (f_vec - self.freq + 1j * self.decay_rate)
        exp_arg /= (self.sigma * np.sqrt(2))
        fit = np.real(wofz(exp_arg)) / self.sigma

        return self.amp * fit / np.max(fit)




"""
lambast/mitigation_methods/density_ratio/weighter.py

Density ratio weighting.
"""

from __future__ import annotations

from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from .datasets import DomainDataset, make_balanced_domain_batch
from .models import DomainClassifier
from .util import _to_float_tensor

TensorLike = Union[np.ndarray, torch.Tensor]


class DensityRatioWeighter:
    """
    Train a domain classifier to distinguish source vs target,
    then compute density-ratio weights for source samples:

        w(x) = d(x) / (1 - d(x))

    where d(x) = P(target | x) from a balanced domain classifier.
    Stores:
      - diagnostics_: summary metrics after fit
      - history_: per-epoch train/val loss & accuracy for the domain classifier
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        *,
        epochs: int = 10,
        batch_size: int = 256,
        lr: float = 1e-3,
        weight_decay: float = 1e-4,
        device: Optional[torch.device] = None,
        eps: float = 1e-4,
        w_clip: Tuple[float, float] = (0.1, 10.0),
        normalize: bool = True,
    ):
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.weight_decay = weight_decay
        self.eps = eps
        self.w_clip = w_clip
        self.normalize = normalize

        self.device = (
            device if device is not None
            else torch.device("cuda" if torch.cuda.is_available() else "cpu")
        )

        self.model = model
        self._fitted = False
        self.diagnostics_: dict[str, float] = {}
        self.history_: dict[str, list[float]] = {}

    # ----------------------------
    # Public API
    # ----------------------------

    def fit(
            self,
            X_source: TensorLike,
            X_target: TensorLike,
            X_source_val=None,
            X_target_val=None) -> "DensityRatioWeighter":
        """
        Train domain classifier on source vs target.
        Parameters
        ----------
        X_source_train: (Ns_train, C, T)
        X_target_train: (Nt_train, C, T)
        X_source_val:   (Ns_val, C, T) optional
        X_target_val:   (Nt_val, C, T) optional
        """

        domain_ds = DomainDataset(X_source, X_target)
        n_source = domain_ds.n_source
        n_target = domain_ds.n_target

        has_val = (X_source_val is not None) and (X_target_val is not None)
        if has_val:
            val_ds = DomainDataset(X_source_val, X_target_val)
        else:
            val_ds = None

        if self.model is None:
            self.model = DomainClassifier(in_channels=domain_ds.Xs.shape[1])

        self.model.to(self.device)

        optimizer = optim.Adam(
            self.model.parameters(),
            lr=self.lr,
            weight_decay=self.weight_decay,
        )

        criterion = nn.BCEWithLogitsLoss()

        half_batch = self.batch_size // 2
        if self.batch_size % 2 != 0:
            raise ValueError(
                "batch_size must be even for balanced domain training.")

        self.model.train()
        self.history_ = {
            "train_loss": [],
            "train_acc": [],
            "val_loss": [],
            "val_acc": []}
        for epoch in range(self.epochs):
            # number of balanced steps per epoch
            steps = max(n_source, n_target) // half_batch
            running_loss = 0.0
            running_acc = 0.0
            for _ in range(steps):
                src_idx = torch.randint(0, n_source, (half_batch,))
                tgt_idx = torch.randint(0, n_target, (half_batch,))

                batch = make_balanced_domain_batch(
                    domain_ds,
                    self.batch_size,
                    source_indices=src_idx,
                    target_indices=tgt_idx,
                    device=self.device,
                )

                optimizer.zero_grad()
                logits = self.model(batch.x)
                eps_ls = 0.05  # try 0.05, 0.1
                y = batch.y * (1.0 - 2 * eps_ls) + eps_ls
                loss = criterion(logits, y)
                loss.backward()
                optimizer.step()

                with torch.no_grad():
                    preds = (torch.sigmoid(logits) > 0.5).float()
                    acc = (preds == batch.y).float().mean()

                running_loss += float(loss.item())
                running_acc += float(acc.item())
            train_loss = running_loss / steps
            train_acc = running_acc / steps
            # optional val evaluation (full-pass)
            if val_ds is not None:
                val_loss, val_acc = self._eval_domain_full(val_ds)
            else:
                val_loss, val_acc = float("nan"), float("nan")

            self.history_["train_loss"].append(train_loss)
            self.history_["train_acc"].append(train_acc)
            self.history_["val_loss"].append(val_loss)
            self.history_["val_acc"].append(val_acc)

        self._fitted = True

        # Basic diagnostics
        ref_ds = val_ds if val_ds is not None else domain_ds
        final_loss, final_acc = self._eval_domain_full(ref_ds)
        self.diagnostics_["domain_loss"] = float(final_loss)
        self.diagnostics_["domain_accuracy"] = float(final_acc)
        self.diagnostics_["used_validation_for_diagnostics"] = float(
            val_ds is not None)

        return self

    def compute_weights(self, X_source: TensorLike, alpha=1) -> torch.Tensor:
        """
        Compute density-ratio weights for source samples.
        """
        if not self._fitted:
            raise RuntimeError("Must call fit() before compute_weights().")

        Xs = _to_float_tensor(X_source).to(self.device)

        assert self.model is not None
        self.model.eval()
        with torch.no_grad():
            logits = self.model(Xs)
            d = torch.sigmoid(logits)

        d = torch.clamp(d, self.eps, 1.0 - self.eps)

        # odds = p(target|x) / p(source|x)
        w = torch.pow(d / (1.0 - d), alpha)

        # clip
        w = torch.clamp(w, self.w_clip[0], self.w_clip[1])

        # normalize to mean 1
        if self.normalize:
            w = w / w.mean()

        return w.detach().cpu()

    # ----------------------------
    # Internal helpers
    # ----------------------------

    def _compute_domain_diagnostics(self, domain_ds: DomainDataset) -> None:
        """
        Compute simple domain accuracy as a sanity check.
        """

        assert self.model is not None
        self.model.eval()
        with torch.no_grad():
            Xs = domain_ds.Xs.to(self.device)
            Xt = domain_ds.Xt.to(self.device)

            logits_s = self.model(Xs)
            logits_t = self.model(Xt)

            ds = torch.sigmoid(logits_s)
            dt = torch.sigmoid(logits_t)

            # predicted target if prob > 0.5
            acc_s = (ds < 0.5).float().mean()
            acc_t = (dt > 0.5).float().mean()
            acc = 0.5 * (acc_s + acc_t)

        self.diagnostics_["domain_accuracy"] = float(acc.cpu())

    def _domain_epoch_eval(self, model, Xs, Xt, device):
        model.eval()
        bce = torch.nn.BCEWithLogitsLoss(reduction="mean")
        with torch.no_grad():
            # Source label 0
            ys = torch.zeros(Xs.shape[0], dtype=torch.float32, device=device)
            # Target label 1
            yt = torch.ones(Xt.shape[0], dtype=torch.float32, device=device)

            ls = bce(model(Xs), ys).item()
            lt = bce(model(Xt), yt).item()
            loss = 0.5 * (ls + lt)

            ps = (torch.sigmoid(model(Xs)) < 0.5).float().mean().item()
            pt = (torch.sigmoid(model(Xt)) > 0.5).float().mean().item()
            acc = 0.5 * (ps + pt)

        model.train()
        return loss, acc

    @torch.no_grad()
    def _eval_domain_full(
            self, domain_ds: DomainDataset) -> Tuple[float, float]:
        """
        Full evaluation on all source and all target examples.

        Returns
        -------
        loss: float (average of source-loss and target-loss)
        acc:  float (average of source-acc and target-acc)
        """
        assert self.model is not None
        self.model.eval()
        bce = nn.BCEWithLogitsLoss(reduction="mean")

        Xs = domain_ds.Xs.to(self.device)
        Xt = domain_ds.Xt.to(self.device)

        ys = torch.zeros(Xs.shape[0], dtype=torch.float32, device=self.device)
        yt = torch.ones(Xt.shape[0], dtype=torch.float32, device=self.device)

        logits_s = self.model(Xs)
        logits_t = self.model(Xt)

        loss_s = bce(logits_s, ys).item()
        loss_t = bce(logits_t, yt).item()
        loss = 0.5 * (loss_s + loss_t)

        ds = torch.sigmoid(logits_s)
        dt = torch.sigmoid(logits_t)

        acc_s = (ds < 0.5).float().mean().item()
        acc_t = (dt > 0.5).float().mean().item()
        acc = 0.5 * (acc_s + acc_t)

        self.model.train()
        return float(loss), float(acc)



import matplotlib.pyplot as plt
import numpy as np
from scipy import stats

from lambast.generate_data import ClaytonCopula


def run_example():

    n = 4  # number of time-series to generate
    t = 1000  # length of time series (arbitrary units)

    # Generate copula sublcass paramter and markovian time series property
    cop = ClaytonCopula(alpha=10, markovian=True)

    # Define the marginal distribution for the time series
    cop.define_marginal(marginal_family="gamma", loc=None, scale=None)

    # Generate the samples
    samples = cop.sample(n, t)

    # Plot them for visualization
    nc = np.min((n, 5))
    nr = int(np.ceil(n / nc))
    plt.figure(figsize=(3 * nc, 2 * nr))
    for nn in range(n):
        plt.subplot(nr, nc, nn + 1)
        plt.plot(samples[nn], c="black")
        plt.title("Sample " + str(nn))
        plt.xlabel("Time (samples)")
        plt.ylabel("Response")
    plt.tight_layout()
    plt.show()

    # Explore other aspects of the copula object, including the joint copula
    # density and generating random and nonrandom variables from the density.

    # Get the density using default arguments
    u, v, p = cop.density()

    # Visualize the density with quantiles
    qs = np.quantile(p, np.linspace(0, 1, 10))

    # Now plot the density with some nice formatting
    fig, ax = plt.subplots()
    CS = ax.contour(u, v, p, levels=qs, cmap="viridis",
                    vmax=qs[-2], vmin=qs[0])
    ax.set_xlabel("U")
    ax.set_ylabel("V")
    ax.set_title(cop.copula_family + " (a = " + str(cop.alpha) + ")")

    def fmt(x):
        """
        Formatter for the contour labels
        """
        s = f"{x:.1f}"
        if s.endswith("0"):
            s = f"{x:.0f}"
        return rf"{s} \%" if plt.rcParams["text.usetex"] else f"{s}"

    ax.clabel(CS, CS.levels, inline=True, fmt=fmt, fontsize=10)

    # Plot some random realizations of the copula given u = 0.7
    n_draws = 100  # number draws to make
    given_u = np.repeat(0.7, n_draws)  # specify which u in [0,1]

    # Generate from the other variable given u
    v_draws = np.array([cop.variable(uu) for uu in given_u])

    # Plot those values in black
    label = f"Conditional Draws: V ~ P(v|u) = {given_u[0]})"
    plt.scatter(given_u, v_draws, facecolors='none', edgecolors='k', s=10,
                label=label)

    # Alternatively, we randomly draw by providing random numbers to "variable"
    u_random = stats.uniform.rvs(0, 1, n_draws)
    w_random = stats.uniform.rvs(0, 1, n_draws)
    v_random = cop.variable(u_random, w_random)
    plt.scatter(u_random, v_random, facecolors='none', edgecolors='r', s=10,
                label="Random Draws")
    plt.legend()
    plt.show()

    # We can verify fundamental properties of the markovian copula with the
    # defined marginal:

    # 1) The time series generated by the copula (u(t)) should have a uniform
    # distribution
    # 2) The time series after passing through the defined marginal
    # distribution should follow the distribution defined.
    # 3) The joint distribution of u(t) and u(t+1) should follow the joint
    # copula distribution.
    # 4) The joint distribution of x(t) and x(t+1) will follow an unspecified,
    # probably complicated distribution that is difficult to characterize.

    # Here, we sample from the copula, but apply no marginal distribution to
    # the samples. i.e. we could set the marginal_family to "uniform", or just
    # pull the uniform_samples object from the copula. This is u(t).

    # Select the index of the sample you want to explore
    nn = 0

    # Pull the samples before they are passed to the marginal distribution
    uniform_samples = cop.uniform_samples
    # the first variable from the copula u(t)
    u_t = uniform_samples[nn][0:t - 1]
    u_tp1 = uniform_samples[nn][1:]  # the second variable u(t+1)

    # x(t) has the marginal distribution applied to it:
    x_t = samples[nn][0:t - 1]  # considered the "time-series" x(t)
    x_tp1 = samples[nn][1:]  # x(t+1)

    nr = 2
    nc = 3
    plt.figure(figsize=(12, 4))

    plt.subplot(nr, nc, 1)
    plt.title("Time Series x(t)")
    plt.plot(x_t, color="k", linewidth=1)
    plt.ylabel("x(t)")
    plt.xlabel("Time (samples)")

    plt.subplot(nr, nc, 2)
    plt.hist(x_t, color="k")
    plt.ylabel("Count")
    plt.xlabel("x(t)")
    plt.title("Distribution of x(t) Follows Marginal (f(x))")
    plt.tight_layout()

    plt.subplot(nr, nc, 3)
    plt.scatter(x_t, x_tp1, s=1, color="k")
    plt.xlabel("x(t)")
    plt.ylabel("x(t+1)")
    plt.title("Joint Distribution")
    plt.tight_layout()

    plt.subplot(nr, nc, 4)
    plt.title("F^-1(x(t)) Follows the Copula")
    plt.scatter(u_t, u_tp1, s=1, color='blue')
    plt.xlabel("F^-1(x(t))")
    plt.ylabel("F^-1(x(t+1))")

    plt.subplot(nr, nc, 5)
    plt.title("Distribution of F^-1(x(t))")
    plt.hist(u_t, color="blue")
    plt.xlabel("F^-1(x(t))")
    plt.ylabel("Count")

    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    run_example()



"""
LAMBAST
=======
Los Alamos Model Bias Assessment and Statistical Toolkit

Subpackages
-----------
detection_methods
    Library of methods for out of distribution or change point detection.
generate_data
    Library of methods to generate distributions
utils
    General utility functions
"""
import importlib as _importlib

from lambast.version import version as __version__

submodules = ["detection_methods", "generate_data", "utils"]
optional = ["__qualname__", "__date__", "__author__", "__credits__"]

__all__ = submodules + ["__version__"]


def __getattr__(name):
    if name in submodules:
        return _importlib.import_module(f'lambast.{name}')
    else:
        try:
            return globals()[name]
        except KeyError:
            if name in optional:
                pass
            else:
                raise ModuleNotFoundError(
                    f"Module {name} does not exist in LAMBAST")



"""
lambast/mitigation_methods/density_ratio/datasets.py

Creates datasets for source and target domains, including
with weights from density ratio estimation.

"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple, Union

import numpy as np
import torch
from torch.utils.data import Dataset

from .util import _to_float_tensor

TensorLike = Union[np.ndarray, torch.Tensor]


class DomainDataset(Dataset):
    """
    Dataset for training a domain classifier: source vs target.

    Exposes two indexing modes:
      - standard indexing: returns (x, domain_y) with domain_y in {0,1}
      - domain-specific indexing via get_source(i) / get_target(i)
        to support balanced batching logic.

    Expected shapes:
      X_source: (N_s, C, T)
      X_target: (N_t, C, T)
    """

    def __init__(
        self,
        X_source: TensorLike,
        X_target: TensorLike,
        return_domain_label: bool = True,
    ):
        self.Xs = _to_float_tensor(X_source)
        self.Xt = _to_float_tensor(X_target)

        if self.Xs.ndim != 3 or self.Xt.ndim != 3:
            raise ValueError(
                "X_source and X_target must be 3D tensors of shape (N, C, T). "
                f"Got X_source.ndim={self.Xs.ndim}"
                f", X_target.ndim={self.Xt.ndim}."
            )
        if self.Xs.shape[1] != self.Xt.shape[1]:
            raise ValueError(
                "Source/target must have same number of channels C. "
                f"Got {self.Xs.shape[1]} vs {self.Xt.shape[1]}."
            )
        if self.Xs.shape[2] != self.Xt.shape[2]:
            raise ValueError(
                "Source/target must have same time length T by default. "
                f"Got {self.Xs.shape[2]} vs {self.Xt.shape[2]}."
            )

        self.return_domain_label = return_domain_label

        self.n_source = int(self.Xs.shape[0])
        self.n_target = int(self.Xt.shape[0])
        self._len = self.n_source + self.n_target

    def __len__(self) -> int:
        return self._len

    def get_source(self, i: int) -> torch.Tensor:
        """Return a single source example x."""
        x = self.Xs[i]
        return x

    def get_target(self, i: int) -> torch.Tensor:
        """Return a single target example x."""
        x = self.Xt[i]
        return x

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Return (x, domain_y) where domain_y is float tensor scalar:
          0.0 for source, 1.0 for target
        """
        if idx < 0 or idx >= self._len:
            raise IndexError(idx)

        if idx < self.n_source:
            x = self.get_source(idx)
            y = torch.tensor(0.0)
        else:
            x = self.get_target(idx - self.n_source)
            y = torch.tensor(1.0)

        if self.return_domain_label:
            return x, y
        # If caller wants only x, still return a consistent tuple
        return x, y


class WeightedTaskDataset(Dataset):
    """
    Dataset for downstream supervised training on SOURCE with sample weights.

    Expected shapes:
      X: (N, C, T)
      y: (N,) or (N,1)  (binary labels typically)
      w: (N,)           (positive weights)
    """

    def __init__(
        self,
        X: TensorLike,
        y: TensorLike,
        w: TensorLike,
    ):
        self.X = _to_float_tensor(X)
        if self.X.ndim != 3:
            raise ValueError(
                "X must be 3D tensor of shape (N, C, T). "
                f"Got X.ndim={self.X.ndim}."
            )

        self.y = _to_float_tensor(y).view(-1)
        self.w = _to_float_tensor(w).view(-1)

        n = int(self.X.shape[0])
        if self.y.shape[0] != n:
            raise ValueError(
                f"y must have length N={n}; got {self.y.shape[0]}.")
        if self.w.shape[0] != n:
            raise ValueError(
                f"w must have length N={n}; got {self.w.shape[0]}.")

    def __len__(self) -> int:
        return int(self.X.shape[0])

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, ...]:
        x = self.X[idx]
        y = self.y[idx]
        w = self.w[idx]
        return x, y, w


@dataclass(frozen=True)
class BalancedDomainBatch:
    """
    Convenience container for a balanced domain batch.
    x: (B, C, T)
    y: (B,) float labels 0/1
    """
    x: torch.Tensor
    y: torch.Tensor


def make_balanced_domain_batch(
    domain_ds: DomainDataset,
    batch_size: int,
    *,
    source_indices: torch.Tensor,
    target_indices: torch.Tensor,
    device: Optional[torch.device] = None,
) -> BalancedDomainBatch:
    """
    Build a balanced domain batch explicitly.
    Caller supplies indices for source and target.

    batch_size must be even; half from source, half from target.
    """
    if batch_size % 2 != 0:
        raise ValueError(
            "batch_size must be even for balanced domain batching.")

    half = batch_size // 2
    if source_indices.numel() != half or target_indices.numel() != half:
        raise ValueError(
            f"Expected {half} source indices and {half} target indices, "
            f"got {source_indices.numel()} and {target_indices.numel()}."
        )

    xs = torch.stack([domain_ds.get_source(int(i))
                     for i in source_indices], dim=0)
    xt = torch.stack([domain_ds.get_target(int(i))
                     for i in target_indices], dim=0)

    x = torch.cat([xs, xt], dim=0)
    y = torch.cat([torch.zeros(half, dtype=torch.float32),
                   torch.ones(half, dtype=torch.float32)], dim=0, )

    # Shuffle within the batch so the model doesn't see ordered domains.
    perm = torch.randperm(batch_size)
    x = x[perm]
    y = y[perm]

    if device is not None:
        x = x.to(device, non_blocking=True)
        y = y.to(device, non_blocking=True)

    return BalancedDomainBatch(x=x, y=y)



"""
Classes to generate non-time series data.
"""

import numpy as np
from numpy.typing import ArrayLike


class BaseDataClass(object):

    def __init__(self) -> None:
        """
        Base class for non-timeseries data.
        """

        return None

    def sample(self, num_samples: int, *args,
               **kwargs) -> ArrayLike | tuple[ArrayLike, ArrayLike]:
        """
        Sample `n` points from the BaseDataClass model.

        Parameters:
        num_samples: number of points to sample
        """
        raise NotImplementedError("Subclasses must implement this method.")


class MultivariateGaussian(BaseDataClass):
    def __init__(self, mean: ArrayLike, covariance: ArrayLike) -> None:
        """
        Initialize the Multivariate Gaussian model.

        Parameters:
        - mean: A 1D numpy array representing the mean vector.
        - covariance: A 2D numpy array representing the covariance matrix.
        """
        super().__init__()

        self.mean = np.array(mean)
        self.covariance = np.array(covariance)

        # Validate covariance matrix
        if self.covariance.shape[0] != self.covariance.shape[1]:
            raise ValueError("Covariance matrix must be square.")
        if self.mean.shape[0] != self.covariance.shape[0]:
            e = "Mean vector size must match covariance matrix dimensions."
            raise ValueError(e)
        if not np.allclose(self.covariance, self.covariance.T):
            raise ValueError("Covariance matrix must be symmetric.")
        if not np.all(np.linalg.eigvals(self.covariance) >= 0):
            e = "Covariance matrix must be positive semi-definite."
            raise ValueError(e)

    def sample(self, num_samples: int = 1) -> ArrayLike:
        """
        Generate samples from the multivariate Gaussian distribution.

        Parameters:
        - num_samples: Number of samples to generate.

        Returns:
        - samples: A 2D numpy array where each row is a sample.
        """
        samples = np.random.multivariate_normal(self.mean, self.covariance,
                                                num_samples)
        return samples


class GaussianMixtureModel(BaseDataClass):
    def __init__(self, weights: ArrayLike, means: list[ArrayLike],
                 covariances: list[ArrayLike]) -> None:
        """
        Initialize the Gaussian Mixture Model.

        Parameters:
        - weights: A list or 1D numpy array of weights for each component
            (should sum to 1).
        - means: A list of 1D numpy arrays representing the mean vector for
            each component.
        - covariances: A list of 2D numpy arrays representing the covariance
            matrix for each component.
        """
        super().__init__()

        self.weights = np.array(weights)
        self.means = [np.array(mean) for mean in means]
        self.covariances = [np.array(cov) for cov in covariances]

        # Validate inputs
        if not np.isclose(np.sum(self.weights), 1):
            raise ValueError("Weights must sum to 1.")
        if len(self.weights) != len(self.means) or \
                len(self.means) != len(self.covariances):

            e = "Number of weights, means, and covariances must be the same."
            raise ValueError(e)

        for mean, cov in zip(self.means, self.covariances):
            # Check if the covariance matrix is square
            if cov.shape[0] != cov.shape[1]:
                raise ValueError("Each covariance matrix must be square.")
            # Check if the dimensions of the mean match the covariance matrix
            if len(mean) != cov.shape[0]:
                e = "The size of the mean vector must match the dimensions of"
                e += " the covariance matrix."
                raise ValueError(e)
            # Check if the covariance matrix is symmetric
            if not np.allclose(cov, cov.T):
                raise ValueError("Each covariance matrix must be symmetric.")
            # Check if the covariance matrix is positive semi-definite
            if not np.all(np.linalg.eigvals(cov) >= 0):
                e = "Each covariance matrix must be positive semi-definite."
                raise ValueError(e)

    def sample(self, num_samples: int = 1) -> tuple[ArrayLike, ArrayLike]:
        """
        Generate samples from the Gaussian Mixture Model.

        Parameters:
        - num_samples: Number of samples to generate.

        Returns:
        - samples: A 2D numpy array where each row is a sample.
        - labels: A 1D numpy array of component labels corresponding to the
            samples.
        """
        n_components = len(self.weights)
        samples = []
        labels = []

        # Generate samples
        for _ in range(num_samples):
            # Select a component based on the weights
            component = np.random.choice(n_components, p=self.weights)
            # Sample from the selected Gaussian component
            sample = np.random.multivariate_normal(self.means[component],
                                                   self.covariances[component])
            samples.append(sample)
            labels.append(component)

        return np.array(samples), np.array(labels)



from .datasets import DomainDataset, WeightedTaskDataset
from .models import BinaryCNN
from .train_task import eval_binary_accuracy, train_binary_classifier
from .weighter import DensityRatioWeighter

__all__ = ["DensityRatioWeighter",
           "DomainDataset",
           "WeightedTaskDataset",
           "BinaryCNN",
           "train_binary_classifier",
           "eval_binary_accuracy"
           ]



"""
lambast/mitigation_methods/density_ratio/models.py

PyTorch models and modules for binary classification and domain classification.

"""

from __future__ import annotations

import torch
import torch.nn as nn


class FeatureNet1D(nn.Module):
    """
    Feature extractor for 1D time series.

    Input:  (B, C, T)
    Output: (B, D)
    """

    def __init__(self, in_channels: int = 2, hidden: int = 32,
                 out_dim: int = 64):
        super().__init__()

        self.conv = nn.Sequential(
            nn.Conv1d(in_channels, hidden, kernel_size=7, padding=3),
            nn.ReLU(),
            nn.Conv1d(hidden, hidden, kernel_size=5, padding=2),
            nn.ReLU(),
            nn.Conv1d(hidden, hidden, kernel_size=3, padding=1),
            nn.ReLU(),
        )

        self.pool = nn.AdaptiveAvgPool1d(1)  # (B, hidden, 1)
        self.proj = nn.Sequential(
            nn.Linear(hidden, out_dim),
            nn.ReLU(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.conv(x)              # (B, hidden, T)
        z = self.pool(z).squeeze(-1)  # (B, hidden)
        z = self.proj(z)              # (B, out_dim)
        return z


class DomainHead(nn.Module):
    """
    Domain classifier head.

    Input:  (B, D)
    Output: (B,) logits
    """

    def __init__(self, in_dim: int = 64):
        super().__init__()
        self.fc = nn.Linear(in_dim, 1)

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        return self.fc(z).squeeze(-1)


class DomainClassifier(nn.Module):
    """
    Full domain classifier: FeatureNet + DomainHead

    Input:  (B, C, T)
    Output: (B,) logits
    """

    def __init__(self, in_channels: int = 2, hidden: int = 32,
                 feat_dim: int = 64):
        super().__init__()
        self.feature = FeatureNet1D(
            in_channels=in_channels,
            hidden=hidden,
            out_dim=feat_dim)
        self.head = DomainHead(in_dim=feat_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.feature(x)
        return self.head(z)


class BinaryCNN(nn.Module):
    """
    Simple 1D CNN for binary classification.
    Input:  (B, C, T)
    Output: (B,) logits
    """

    def __init__(self, in_channels: int = 2, hidden: int = 32):
        super().__init__()
        self.net = nn.Sequential(
            nn.Conv1d(in_channels, hidden, kernel_size=7, padding=3),
            nn.ReLU(),
            nn.Conv1d(hidden, hidden, kernel_size=5, padding=2),
            nn.ReLU(),
            nn.Conv1d(hidden, hidden, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.AdaptiveAvgPool1d(1),
        )
        self.head = nn.Linear(hidden, 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.net(x).squeeze(-1)      # (B, hidden)
        return self.head(z).squeeze(-1)  # (B,)



"""
Utility functions.

"""

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray


def assert_valid_covariance(matrix: NDArray, tol: float = 1e-8) -> None:
    """
    Raises an error if a matrix is not a valid covariance matrix.

    Parameters:
        matrix array: The matrix to check.
        tol (float): Tolerance for checking symmetry and non-negative
            eigenvalues.
    """

    # Check if the matrix is square
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Covariance matrix is not square.")

    # Check symmetry
    if not np.allclose(matrix, matrix.T, atol=tol):
        raise ValueError("Covariance matrix is not symmetric.")

    # Check positive semi-definiteness using eigenvalues
    eigenvalues = np.linalg.eigvals(matrix)
    if np.any(eigenvalues < -tol):
        raise ValueError("Covariance matrix is not positive semi-definite.")


def compute_snr(sig: NDArray[np.complex64],
                noise: NDArray[np.complex64]) -> NDArray[np.complex64]:
    """
    Compute SNR empirically given signal and noise data.
    """

    noise_var = np.var(noise - np.mean(noise, 1, keepdims=True), 1)
    sig_var = np.var(sig - np.mean(sig, 1, keepdims=True), 1)

    return 10.0 * np.log10(sig_var / noise_var)


def white_noise_gen(t: NDArray, N: int, sigma: float = 1.0,
                    rng: np.random.Generator = np.random.default_rng()
                    ) -> NDArray:
    """
    Generate complex Gaussian white noise time series of shape (N, t) with
    total variance sigma.
    """

    r_dist1 = rng.normal(0, scale=sigma / np.sqrt(2), size=(N, len(t)))
    i_dist2 = rng.normal(0, scale=sigma / np.sqrt(2), size=(N, len(t)))

    return r_dist1 + 1j * i_dist2


def split_noise(noise, n_timesteps):
    """
    Split noise data array along time axis to generate a higher number of
    shorter examples
    """

    # TEST: does this work for an edge case
    # where (noise.shape[1] % n_timesteps) == 0?
    nseg = noise.shape[1] // n_timesteps
    noise = noise[:, : (nseg * n_timesteps)]
    snoise = np.stack(np.hsplit(noise, nseg), axis=1)
    rsnoise = np.reshape(snoise, (noise.shape[0] * nseg, -1))

    return rsnoise


def plot_complex_ts(t, y, ax=None, **kwargs):
    """
    Plot complex-valued time series
    """

    if ax is None:
        ax = plt.gca()
    for f, c in zip([np.real, np.imag], ["k", "r"]):
        ax.plot(t, f(y), c, **kwargs, label=f.__name__)



import warnings

import numpy as np
from numpy.typing import NDArray
from scipy import stats
from scipy.optimize import fsolve

from .timeseries_class import TimeSeries


class Copula(TimeSeries):

    def __init__(self, alpha: float | None = None,
                 markovian: bool = True) -> None:
        """
        Initializes a new instance of a Copula.

        Parameters:
            alpha: parameter controlling the copula, currently only
                archimedian copulas are implemented
            markovian: default True, if True, then the copula samples have a
                markovian process property, i.e. correlated in time.
        The functions here-in rely heavily on Chapter 2 of the book:
            Sun, Li-Hsien, et al. Copula-Based Markov Models for Time Series:
            Parametric Inference and Process Control,
            Springer Singapore Pte. Limited, 2020.
        """
        super().__init__()
        self.alpha = alpha
        self.markovian = markovian

        # Family that will be defined by instance of subclass
        self.copula_family: str | None = None

    def define_marginal(self, marginal_family: str = "uniform",
                        loc: float | None = None,
                        scale: float | None = None) -> None:
        """
        Method to instantiate the desired marginal distribution qualities.

        Arguments:
            marginal_family: string, choices include:
                "uniform", "normal", "gamma", "t", "gumbel", "exponential"
            loc: location parameter for distributions that accept them for
                exponential, loc corresponds to rate for t, loc corresponds
                to degrees of freedom
            scale: scale parameter for distributions that accept them
        """
        self.marginal_family = marginal_family
        self.loc = loc
        self.scale = scale

    def variable_generator(self, u: NDArray, w: NDArray) -> NDArray:
        """
        The variable_generator used for the copula model to produce
        realizations from its PDF.

        Parameters:
            u: number between (0,1)
            w: number between (0,1)
        Returns v such that (U,V) is a realization of the copula model.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def variable(self, u: NDArray | None = None, w: NDArray | None = None,
                 rng: np.random.Generator | None = None,
                 *args) -> float | NDArray:
        """
        Provides variables u and w to the copula variable_generator and returns
        the variable v such that (U,V) is a realization of the copula model.

        If u and w are random or None, then provides a random variable.

        Parameters:
            u: uniform variable between (0,1)
            w: uniform variable between (0,1)
        """
        if u is None:
            u = stats.uniform.rvs(0, 1, random_state=rng)
        if w is None:
            w = stats.uniform.rvs(0, 1, random_state=rng)

        return self.variable_generator(u, w)

    def sample(self, n: int = 1, t: int = 1000,
               rng: np.random.Generator | None = None) -> NDArray:
        """
        Generate n samples from a bivariate copula distribution with marginal
        family specified.

        If markovian, x_t and x_t+1 will be distributed according to the
        copula. The marginal distribution will be uniform.

        Parameters:
            n: number of samples to draw of length t
            t: number of time points
            rng: optional, the random number generator
        Returns ndarray of random variables of shape [n, t]
        """
        # Initialize empty matrix
        uniform_samples = np.zeros((n, t))

        for nn in range(n):
            # Random initialization (get the chain going)
            u = stats.uniform.rvs(0, 1, random_state=rng)
            for tt in range(t):
                # Do not provide w to get a random sample
                v = self.variable(u=u, w=None, rng=rng)
                uniform_samples[nn, tt] = v

                if self.markovian:
                    u = v

        self.uniform_samples = uniform_samples
        self.samples = self.uniform_to_marginal(self.uniform_samples)

        return self.samples

    def density_generator(self, u: NDArray, v: NDArray) -> NDArray:
        """
        The joint density of the bivariate copula

        Arguments:
            u: value between (0,1)
            v: value between (0,1)
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def density(self, u: NDArray | None = None, v: NDArray | None = None,
                n_samples: int = 2000) -> tuple[NDArray, NDArray, NDArray]:
        """
        Function to call the density_generator function specific to each
        copula. If no arguments are given, then provides a meshgrid that can
        easily be passed to plt.contour()

        Arguments:
            u: array
            v: array of same size as u
            n_samples: number of samples to draw if neither u or v are given
            Must provide both u and v or neither u and v.
        Returns u, v, p with p being the density values
        """

        if u is None and v is None:
            u_vec = np.linspace(0.001, 0.999, n_samples)
            v_vec = np.linspace(0.001, 0.999, n_samples)
            u, v = np.meshgrid(u_vec, v_vec)

        if u is None or v is None:
            e = "Must provide both u and v or neither u and v."
            raise ValueError(e)

        p = self.density_generator(u, v)

        return u, v, p

    def uniform_to_marginal(self, uniform_samples: NDArray) -> NDArray:
        """
        Convert the data to the uniform marginals to have a marginal
        distribution of interest.

        Generates random variables from uniform RVs using inverse CDF from
        distribution specified by family. Updates class with valid loc and
        scale parameters if None was provided.

        Arguments:
            uniform_samples: array of samples between (0,1)
        Returns samples after being passed through the marginal distribution.
        """

        match self.marginal_family:
            case "uniform":
                return uniform_samples  # do nothing, keep marginal uniform
            case "normal":
                if self.loc is None:
                    self.loc = 0
                if self.scale is None:
                    self.scale = 1
                return stats.norm.ppf(uniform_samples, self.loc, self.scale)
            case "gamma":
                if self.loc is None:
                    self.loc = 1
                if self.scale is None:
                    self.scale = 2
                return stats.gamma.ppf(uniform_samples, self.loc, self.scale)
            case "t":
                if self.loc is None:
                    self.loc = 3  # really degrees of freedom
                return stats.t.ppf(uniform_samples, self.loc)
            case "gumbel":
                if self.loc is None:
                    self.loc = 3
                if self.scale is None:
                    self.scale = 4
                return stats.gumbel_r.ppf(uniform_samples, self.loc,
                                          self.scale)
            case "exponential":
                if self.loc is None:
                    self.loc = 1  # rate
                return stats.expon.ppf(uniform_samples, self.loc)

        raise ValueError(
            "Marginal family chosen is not supported." +
            " Please see possible options."
        )


class ClaytonCopula(Copula):
    """
    Clayton: lower tail dependence (strong dependence at (U, V)≈(0,0)). The
    level of dependence increases as α increases. bounds: [-1, inf). This code
    does not support alpha = -1

    "The Clayton and Frank copulas converge to the independence copula
    when α → 0."

    "The Clayton copula allows negative dependence for α ∈ [−1, 0),
    particularly giving the Fréchet–Hoeffding lower bound copula for α = −1.

    However, the copula density does not exist on the line u−α+v−α−1 = 0,
    giving a “singular” distribution for α ∈ [−1, 0).

    Hence, the Clayton copula is usually not suitable for modeling negative
    dependence except for some special examples (e.g., Emura et al. 2011)."
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.copula_family = "Clayton"

        assert self.alpha is not None
        if (self.alpha <= -1) | (self.alpha == 0):
            e = "Parameter bounds may be [-1, inf) and not 0 for Clayton "
            e += "copula. This code does not support alpha = -1"
            raise ValueError(e)

    def variable_generator(self, u: NDArray, w: NDArray) -> NDArray:

        assert self.alpha is not None
        v = ((w ** (-self.alpha / (self.alpha + 1)) - 1)
             * u ** (-self.alpha) + 1) ** (-1 / self.alpha)

        return v

    def density_generator(self, u: NDArray, v: NDArray) -> NDArray:
        # Bivariate density
        assert self.alpha is not None
        p = (1 + self.alpha) * (u * v) ** (-self.alpha - 1)
        p *= (u ** -self.alpha + v ** -self.alpha - 1) ** (-1 / self.alpha - 2)

        return p


class JoeCopula(Copula):
    """
    Joe:  upper tail dependence (strong dependence at (U, V)≈(1,1)). The level
    of dependence increases as α increases.

    "The Joe copula becomes the independent copula when α = 1." - Copula book
    Getting these variables requires a somewhat finicky solver/optimization.
    bounds: [1, inf).
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.copula_family = "Joe"

        assert self.alpha is not None
        if self.alpha < 1:
            e = "Parameter bounds may be [1, inf) for Joe copula"
            raise ValueError(e)
        if self.alpha > 200:
            w = "Large values of alpha (>200) can cause instability in the "
            w += "number generation and it is recommended to keep alpha below "
            w += "this value."
            warnings.warn(w)

    def solver(self, v: NDArray, u: NDArray, w: NDArray) -> NDArray:
        """
        Solver function for the joe copula.
        Parameters:
            v: array of places to initialize solver
            u: uniform variable
            w: uniform variable
        """

        assert self.alpha is not None
        term1 = (1 - u) ** self.alpha + (1 - v) ** self.alpha
        term1 -= ((1 - u) * (1 - v)) ** self.alpha
        term1 **= 1 / self.alpha - 1

        term2 = (1 - (1 - v) ** self.alpha) * (1 - u) ** (self.alpha - 1)

        return term1 * term2 - w

    def variable_generator(self, u: NDArray, w: NDArray) -> NDArray:

        # First, get a good spot for initializing the solver
        ntry = 1000
        vtry = np.linspace(0.01, 0.99, ntry)
        min_index = np.argmin(self.solver(vtry, u, w) ** 2)

        # Initialization
        v0 = vtry[min_index]
        roots = fsolve(self.solver, v0, (u, w))

        if len(roots) > 1:
            print("Multiple roots, troubleshoot solver")

        return roots[0]

    def density_generator(self, u: NDArray, v: NDArray) -> NDArray:

        # Precalculate common factors
        assert self.alpha is not None
        product = (1 - u) * (1 - v)
        common = (1 - u) ** self.alpha + (1 - v) ** self.alpha
        common -= product ** self.alpha

        term1 = self.alpha * common ** (1 / self.alpha - 1)
        term1 *= product ** (self.alpha - 1)

        term2 = (self.alpha - 1) * common ** (1 / self.alpha - 2)

        term3 = (1 - (1 - u) ** self.alpha) * (1 - (1 - v) ** self.alpha)
        term3 *= product ** (self.alpha - 1)

        # Bivariate density
        return term1 + term2 * term3


class FrankCopula(Copula):
    """
    Frank: A negative value of α gives negative dependence while a positive
    value of α gives positive dependence between U and V.

    Getting these variables requires a somewhat finicky solver/optimization.
    bounds: [-inf, 0) or (0, inf].

    The Clayton and Frank copulas converge to the independence
    copula when α → 0.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.copula_family = "Frank"

        assert self.alpha is not None
        if self.alpha == 0:
            e = "Parameter bounds may be (-inf, inf) but not 0 for "
            e += "Frank copula"
            raise ValueError(e)
        if self.alpha > 500:
            w = "Large values of alpha (>500) can cause instability in the "
            w += "number generation and it is recommended to keep alpha "
            w += "below 500."
            warnings.warn(w)

    def variable_generator(self, u: NDArray, w: NDArray) -> NDArray:

        assert self.alpha is not None
        term1 = -1 / self.alpha
        lognum = np.exp(-self.alpha * u) - w * \
            np.exp(-self.alpha * u) + w * np.exp(-self.alpha)

        if lognum == 0:
            print(u, w, self.alpha)

        logden = np.exp(-self.alpha * u) - w * np.exp(-self.alpha * u) + w

        return term1 * (np.log(lognum) - np.log(logden))

    def density_generator(self, u: NDArray, v: NDArray) -> NDArray:

        assert self.alpha is not None
        num = self.alpha * (1 - np.exp(-self.alpha)) * \
            np.exp(-self.alpha * (u + v))
        den = (np.exp(-self.alpha) - 1 + (np.exp(-self.alpha * u) - 1)
               * (np.exp(-self.alpha * v) - 1))**2

        # Bivariate density
        return num / den


class NormalCopula(Copula):
    """
    Normal: the parameter specified controls the correlation, bounds: (-1,1)
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.copula_family = "Normal"

        assert self.alpha is not None
        if np.abs(self.alpha) > 1:
            e = "Parameter bounds may be (-1,1) for Normal copula"
            raise ValueError(e)

    def variable_generator(self, u: NDArray, w: NDArray) -> NDArray:
        psi_i_k = stats.norm.ppf(w)
        psi_i_u = stats.norm.ppf(u)

        assert self.alpha is not None
        inner_term = psi_i_k * \
            np.sqrt(1 - self.alpha ** 2) + self.alpha * psi_i_u
        v = stats.norm.cdf(inner_term)
        return v

    def density_generator(self, u: NDArray, v: NDArray) -> NDArray:
        psi_i_u = stats.norm.ppf(u)
        psi_i_v = stats.norm.ppf(v)

        assert self.alpha is not None
        term1 = 1 / np.sqrt(1 - self.alpha**2)

        # NOTE: in the following term, the book has a sign wrong, this has
        # been corrected here:
        exp_num = - self.alpha**2 * \
            (psi_i_u**2 + psi_i_v**2) + 2 * self.alpha * psi_i_u * psi_i_v
        exp_den = 2 * (1 - self.alpha**2)

        # Bivariate density
        return term1 * np.exp(exp_num / exp_den)



"""
Initial shift examples.

"""
import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm

from lambast.generate_data import LinearSSM


def avg_acf(ts):
    """
    Helper function to visualize average acf
    """
    return np.mean([sm.tsa.acf(x) for x in ts], axis=0)


def plot_ts_and_new_ts(ts, new_ts):
    """
    Factor out the comparison plots
    """
    fig = plt.figure(figsize=(10, 8))
    grid = fig.add_gridspec(2, 2, width_ratios=[2, 1])

    # Create the subplots
    ax1 = fig.add_subplot(grid[0, 0])  # Top left (spans 2 columns)
    ax2 = fig.add_subplot(grid[0, 1])   # Top right
    ax3 = fig.add_subplot(grid[1, 0])  # Bottom left (spans 2 columns)
    ax4 = fig.add_subplot(grid[1, 1])   # Bottom right

    ax1.plot(ts[:, 1, :].T)
    ax1.set_title('Training data')

    ax2.plot(avg_acf(ts[:, 1, :]))
    ax2.set_title('Training autocorr.')

    ax3.plot(new_ts[:, 1, :].T)
    ax3.set_title('Shifted data')

    ax4.plot(avg_acf(new_ts[:, 1, :]))
    ax4.set_title('Shifted autocorr.')

    plt.tight_layout()
    plt.show()


def run_example():
    """
    Driver function to run this example

    """
    # Settings
    rng = np.random.default_rng(seed=42)
    d = 10
    p = 2
    n = 10
    t = 150

    # Generate "training data"
    state_matrix = rng.normal(size=(d, d))
    obs_matrix = rng.normal(size=(p, d))
    state_noise_cov = 0.1 * np.eye(d)
    obs_noise_cov = 0.01 * np.eye(p)

    # Instantiate the SSM object
    ssm = LinearSSM(state_matrix, state_noise_cov, obs_matrix, obs_noise_cov,
                    rng=rng, scale_matrix=True)

    # Sample the time series
    ssm.sample(n, t)
    ssm.plot_sample()

    # Case A: Generate with changed state matrix
    state_matrix_A = ssm.state_matrix + 0.1 * rng.normal(size=(d, d))
    ssm_A = ssm.copy_with_changes(state_matrix=state_matrix_A,
                                  scale_matrix=True)

    ssm_A.sample(n, t)
    plot_ts_and_new_ts(ssm.ts_samples, new_ts=ssm_A.ts_samples)

    # Case B: change observation noise
    obs_noise_cov_B = 1.0 * np.eye(p)
    ssm_B = ssm.copy_with_changes(obs_noise_cov=obs_noise_cov_B)

    ssm_B.sample(n, t)
    plot_ts_and_new_ts(ssm.ts_samples, new_ts=ssm_B.ts_samples)

    # Case C: change observation matrix
    obs_matrix_C = ssm.obs_matrix + 1.5 * rng.normal(size=(p, d))
    ssm_C = ssm.copy_with_changes(obs_matrix=obs_matrix_C)

    ssm_C.sample(n, t)
    plot_ts_and_new_ts(ssm.ts_samples, new_ts=ssm_C.ts_samples)

    # TODO
    # Case D: change point (sort of)


if __name__ == "__main__":
    run_example()



from .copulas import ClaytonCopula, FrankCopula, JoeCopula, NormalCopula
from .generate_non_time_series_data import (GaussianMixtureModel,
                                            MultivariateGaussian)
from .generate_timeseries import HSMM, LinearSSM
from .generate_voigt_signal_data import Voigt, VoigtSignal



import copy
from dataclasses import dataclass

import numpy as np
import scipy as sp
from numpy.typing import ArrayLike, NDArray


@dataclass
class PermutationDistance(object):
    """
    Class that aggregates all detection methods
    Parameters:
    - train_data: A list of training data samples.
    - target_data: A list of target data samples.
    - p_value: P value; default is 0.05.
    - num_samples: Number of points to sample for estimating distribution;
        default is 500.
    - num_bins: Number of bins for binning the data for calculating PSI;
        default is 20.
    - min_bin_prob: Minimum probability values used for a bin to avoid
        issues with empty bin (divide by 0 issues); default is 1*10^-10.
    - n_resamples: Number of permutations evaluated for the permutation
        test; default is 1,000.
    - diff_tol: Difference tolerance to consider data different for
        memoization purposes
    """

    train_data: NDArray | list[float]
    target_data: NDArray | list[float]

    num_bins: int = 20
    num_samples: int = 500
    p_value: float = 5e-2
    min_bin_prob: float = 1e-10
    n_resamples: int = 1000
    wass_p: int = 1
    diff_tol: float = 1e-14

    def __post_init__(self) -> None:
        """
        These are variables used for resampling memoization. That way both
        "self.train_data" and "self.target_data" can be kept unchanged from
        one reshufling to the next. They also help to recognize when we cannot
        rely on memoization and have to re-caculate the expensive values again.
        """

        self.train_d: NDArray | None = None
        self.target_d: NDArray | None = None
        self.prev_train: NDArray | None = None
        self.prev_target: NDArray | None = None

    def __data_range(self, train_data: ArrayLike | None = None,
                     target_data: ArrayLike | None = None
                     ) -> tuple[float, float]:
        """
        Retreive the data range

        Parameters:
        - train_data: A list of training data samples. Default value is given
            at initialization.
        - target_data: A list of target data samples. Default value is given
            at initialization.
        """

        if train_data is None:
            train_data = self.train_data
        if target_data is None:
            target_data = self.target_data

        # Get maxima and minima of data
        all_data = np.concatenate((train_data, target_data))
        return np.min(all_data), np.max(all_data)

    def __get_percent(self, train_data: ArrayLike | None = None,
                      target_data: ArrayLike | None = None,
                      num_bins: int | None = None,
                      min_bin_prob: float | None = None
                      ) -> tuple[NDArray, NDArray]:
        """
        Common operations for histogram calculation

        Parameters:
        - train_data: A list of training data samples. Default value is given
            at initialization.
        - target_data: A list of target data samples. Default value is given
            at initialization.
        - num_bins: Number of bins for binning the data for calculating PSI;
            default is 20.
        - min_bin_prob: Minimum probability values used for a bin to avoid
            issues with empty bin (divide by 0 issues); default is 1e-10.
        """

        if train_data is None:
            train_data = self.train_data
        if target_data is None:
            target_data = self.target_data
        if num_bins is None:
            num_bins = self.num_bins
        if min_bin_prob is None:
            min_bin_prob = self.min_bin_prob

        min_, max_ = self.__data_range(train_data=train_data,
                                       target_data=target_data)

        # Train data percent
        train_p = np.histogram(train_data, bins=num_bins,
                               range=(min_, max_))[0]
        train_p /= np.size(train_data)
        train_p = np.clip(train_p, a_min=min_bin_prob, a_max=None)

        # Target data percent
        target_p = np.histogram(target_data, bins=num_bins,
                                range=(min_, max_))[0]
        target_p /= np.size(target_data)
        target_p = np.clip(target_p, a_min=min_bin_prob, a_max=None)

        return train_p, target_p

    def __get_distrib(self, train_data: ArrayLike | None = None,
                      target_data: ArrayLike | None = None,
                      num_samples: int | None = None
                      ) -> tuple[NDArray, NDArray]:
        """
        Common operations for kde calculation

        Parameters:
        - train_data: A list of training data samples. Default value is given
            at initialization.
        - target_data: A list of target data samples. Default value is given
            at initialization.
        - num_samples: Number of points to sample for estimating distribution;
            default=1000.
        """

        if train_data is None:
            train_data = self.train_data
        if target_data is None:
            target_data = self.target_data
        if num_samples is None:
            num_samples = self.num_samples

        min_, max_ = self.__data_range(train_data=train_data,
                                       target_data=target_data)

        vals = np.linspace(min_, max_, num_samples)

        # Train data distribution
        train_d = sp.stats.gaussian_kde(train_data).evaluate(vals)

        # Target data distribution
        target_d = sp.stats.gaussian_kde(target_data).evaluate(vals)

        return train_d, target_d

    def metric(self, type_: str, use: str,
               train_data: ArrayLike | None = None,
               target_data: ArrayLike | None = None,
               num_bins: int | None = None,
               min_bin_prob: float | None = None,
               num_samples: int | None = None,
               wass_p: float | None = None,
               memoize: bool = False) -> float:
        """
        Returns a given metric

        Parameters:
        - type_: String, metric type, values: "PSI", "JS", "WD", "KS"
        - use: String, either "histogram" or "kde"
        - train_data: A list of training data samples. Default value is given
            at initialization.
        - target_data: A list of target data samples. Default value is given
            at initialization.
        - num_samples: Number of points to sample for estimating distribution;
            default value is given at initialization.
        - num_bins: Number of bins for binning the data for calculating PSI;
            default value is given at initialization.
        - min_bin_prob: Minimum probability values used for a bin to avoid
            issues with empty bin (divide by 0 issues); default value is given
            at initialization.
        - wass_p: Wasserstein distance order; default default value is given
            at initialization.
        - memoize: Wether to memoize or not, defaults to False
        """

        assert type_ in ["PSI", "JS", "WD", "KS"]
        assert use in ["histogram", "kde"]

        if train_data is None:
            train_data = self.train_data
        if target_data is None:
            target_data = self.target_data

        need_dist_estimates = type_ in ["PSI", "JS"]

        if need_dist_estimates:
            # need to calculate distribution estimates
            if not memoize:
                self.train_d = None
                self.target_d = None
            else:
                # Here we are checking if the values have changed, even if
                # asked to memoize, the values should be calculated again if
                # anything has changed.
                #
                # If we do not have previous data, always calculate
                if self.prev_train is None:
                    self.train_d = None
                elif (abs(self.prev_train - train_data) > self.diff_tol).any():
                    self.train_d = None

                # Same as before but with the target data
                if self.prev_target is None:
                    self.target_d = None
                elif (abs(self.prev_target - target_data) >
                      self.diff_tol).any():
                    self.target_d = None

                # Remember the new data
                assert type(target_data) is np.ndarray
                assert type(train_data) is np.ndarray
                self.prev_target = copy.copy(target_data)
                self.prev_train = copy.copy(train_data)

            # Train and target percentages or distributions
            if self.train_d is None or self.target_d is None:
                if use == "histogram":
                    tup = self.__get_percent(train_data=train_data,
                                             target_data=target_data,
                                             num_bins=num_bins,
                                             min_bin_prob=min_bin_prob)
                elif use == "kde":
                    tup = self.__get_distrib(train_data=train_data,
                                             target_data=target_data,
                                             num_samples=num_samples)
                else:
                    e = "Provide a use parameter: 'histogram' or 'kde'"
                    raise Exception(e)

                self.train_d, self.target_d = tup

        assert self.train_d is not None and self.target_d is not None

        match type_:
            case "PSI":
                return np.sum((self.train_d - self.target_d) *
                              np.log(self.train_d / self.target_d))
            case "JS":
                return sp.spatial.distance.jensenshannon(self.train_d,
                                                         self.target_d,
                                                         base=np.e)
            case "WD":
                if wass_p is None:
                    wass_p = self.wass_p
                return np.mean(
                    np.abs(
                        np.sort(train_data) - np.sort(target_data)
                    ) ** wass_p) ** (-wass_p)
            case "KS":
                return sp.stats.kstest(train_data,  # type: ignore[arg-type]
                                       target_data,  # type: ignore[arg-type]
                                       alternative='two-sided').statistic
            case _:
                e = "Provide a type_ parameter: 'PSI', 'JS', 'WD', or 'KS'"
                raise Exception(e)

    def data_shift_test(self, use: str, metrics: list[str] | str,
                        wass_p: float | None = None,
                        num_bins: int | None = None,
                        min_bin_prob: float | None = None,
                        num_samples: int | None = None,
                        n_resamples: int | None = None,
                        p_value: float | None = None
                        ) -> list[tuple[float, bool]] | tuple[float, bool]:
        """
        Return tuple of p_value and boolean value equal to 1 if data shift is
        detected from "self.train_data" to "self.target_data" based on metric
        and permutation test, else 0.

        If list of metrics was given, return list of the tuples described above
        for each type of metric in the same order.

        Parameters:
        - use: String, either "histogram" or "kde"
        - metrics: String or list of strings, metric type, values: "PSI", "JS",
            "WD", "KS"
        - wass_p: Wasserstein distance order; default is 1
        - num_bins: Number of bins for binning the data for calculating PSI;
            default is 20.
        - min_bin_prob: Minimum probability values used for a bin to avoid
            issues with empty bin (divide by 0 issues); default is 1*10^-10.
        - num_samples: Number of points to sample for estimating distribution;
            default is 1000.
        - n_resamples: Number of permutations evaluated for the permutation
            test; default is 100,000.
        - p_value: P value; default is 0.05.
        """

        assert use in ["histogram", "kde"]

        if n_resamples is None:
            n_resamples = self.n_resamples
        if p_value is None:
            p_value = self.p_value
        if wass_p is None:
            wass_p = self.wass_p

        is_list = type(metrics) is list

        if is_list:
            metrics_list = metrics
        else:
            assert type(metrics) is str
            metrics_list = [metrics]

        observed_stat: list[float] = []
        permuted_stats: list[list[float]] = []
        results: list[tuple[float, bool]] = []

        # Save the combined data here
        combined_data = np.concatenate((self.train_data, self.target_data))

        for type_ in metrics_list:

            observed_stat.append(0.)
            permuted_stats.append([])
            results.append((0., False))

            if type_ == "KS":
                ks_test_result = sp.stats.kstest(self.train_data,
                                                 self.target_data,
                                                 alternative='two-sided')

                # Give the results already
                results[-1] = (ks_test_result.pvalue,
                               ks_test_result.pvalue < p_value)
                continue

            observed_stat[-1] = self.metric(type_, use, wass_p=wass_p,
                                            num_bins=num_bins,
                                            min_bin_prob=min_bin_prob,
                                            num_samples=num_samples)

        for i in range(n_resamples):
            np.random.shuffle(combined_data)
            perm_group_a = combined_data[:len(self.train_data)]
            perm_group_b = combined_data[len(self.train_data):]

            for j, type_ in enumerate(metrics_list):
                # Ignore KS, we have that already
                if type_ == "KS":
                    continue

                permuted_stats[j].append(self.metric(type_, use,
                                                     train_data=perm_group_a,
                                                     target_data=perm_group_b,
                                                     wass_p=wass_p,
                                                     num_bins=num_bins,
                                                     min_bin_prob=min_bin_prob,
                                                     num_samples=num_samples,
                                                     memoize=True))

        for j, type_ in enumerate(metrics_list):
            if type_ == "KS":
                continue

            test_p_value = np.mean(
                np.abs(permuted_stats[j]) >= np.abs(observed_stat[j]))
            result = (test_p_value, test_p_value < p_value)
            results[j] = result

        if is_list:
            return results
        else:
            return results[0]
